#include "clothoid.h"

float Clothoid::CURV_FACTOR = sqrtf(M_PI / 2.0f);
float Clothoid::SQRT_2 = sqrtf(2.0f);
float Clothoid::MIN_CURV_DIFF = 0.005f;
float Clothoid::MIN_CURV_CIRCLE = 0.002f;

Clothoid::Clothoid() :
        m_startCurv(0.0f),
        m_endCurv(0.0f),
        m_arcLen(1.0f)
{
}


QDebug operator<<(QDebug d, const Clothoid & c)
{
    d << "startcurv: " << c.m_startCurv << " endCurv: " << c.m_endCurv << " arcLen: " << c.m_arcLen << " B " << c.ComputeB() << " t1 " << c.ComputeT1(c.ComputeB()) << " t2 " << c.ComputeT2(c.ComputeB()) << " endtanangle: " << c.EndTanAngle();
    return d;
}

void Clothoid::DrawGL() const
{

    glBegin(GL_LINE_STRIP);

    for (int i=0; i<m_samples.size(); ++i)
        glVertex2f(m_samples[i].x(), m_samples[i].y());

    glEnd();

}

void Clothoid::DrawArcLengthGL(const float s1, const float s2) const
{

    const int start = qMax(int((s1 / m_arcLen) * m_samples.size()), 0);
    const int end = qMin(int((s2 / m_arcLen) * m_samples.size()), m_samples.size());

    glBegin(GL_LINE_STRIP);

    for (int i=start; i<end; ++i)
        glVertex2f(m_samples[i].x(), m_samples[i].y());

    glEnd();

}

QVector2D Clothoid::Sample(int t) const
{
    return m_samples[t];
}

QVector2D Clothoid::EndPoint() const
{

    if (m_arcLen <= 0.0f)
        return QVector2D(0, 0);

    return m_samples.last();

}

float Clothoid::EndTanAngle() const
{

    if (m_arcLen <= 0.0f)
        return 0.0f;

    if (fabs(m_endCurv - m_startCurv) > MIN_CURV_DIFF) { //clothoid

        const float B = ComputeB();
        const float t1 = ComputeT1(B);
        const float t2 = ComputeT2(B);

        //gotta integrate curvature here for tan angle
        //both integrated individually (always positive)
        float angle1 = t1 * t1;
        float angle2 = t2 * t2;

        //now they may cancel out one another, or add to each other
        float angle = fabs(angle2) - fabs(angle1);

        if ((fabs(t2) > fabs(t1) && (t2 < 0.0f)) ||
            (fabs(t1) > fabs(t2) && (t1 > 0.0f)))
            angle = -angle;

        return angle;
    }
    else if (fabs(m_startCurv) > MIN_CURV_CIRCLE && fabs(m_endCurv) > MIN_CURV_CIRCLE) { //circle

        float circRad = 1.0f / ((m_endCurv + m_startCurv) * 0.5f);
        return m_arcLen / circRad;

    }
    else //line segment
        return 0.0f;

}

float Clothoid::ComputeB() const
{
    return sqrtf(m_arcLen / (M_PI * fabs(m_endCurv - m_startCurv)));
}

float Clothoid::ComputeT1(const float B) const
{
    return m_startCurv * B * CURV_FACTOR;
}

float Clothoid::ComputeT2(const float B) const
{
    return m_endCurv * B * CURV_FACTOR;
}

void Clothoid::DoSamples()
{   

    const int nSamples = int(ceilf(m_arcLen * SAMPLES_PER_UNIT)) + 1;

    m_samples.clear();

    if (fabs(m_endCurv - m_startCurv) < MIN_CURV_DIFF) { //curvature not changing much - line or circle segment


        if (fabs(m_endCurv) < MIN_CURV_CIRCLE || fabs(m_startCurv) < MIN_CURV_CIRCLE) { //a line segment

            //(the radial curvature is very tiny, or the distance to osculating circle is HUGE)
            for (int i=0; i<nSamples; ++i) {

                float interp = float(i) / float(nSamples - 1);
                m_samples.push_back(QVector2D(m_arcLen * interp, 0.0f));

            }

        }
        else {

            float circRad = 1.0f / ((m_endCurv + m_startCurv) * 0.5f);

            //angle sweep should be of length of m_arcLen
            //circumference of circle is 2*pi*r
            //we want an angle that goes of length m_arcLen
            //therefore sweeping angle we want is, if 2*pi does distance 2*pi*r
            //angle = 0 len = 0 angle = 2*pi len = 2*pi*r
            //length of sweep is r times the angle in radians
            // angle should be m_arcLen / circRad right?
            //its late...

            float deg270 = M_PI * 3.0f / 2.0f;
            for (int i=0; i<nSamples; ++i) {

                float interp = (m_arcLen / circRad) * float(i) / float(nSamples - 1) + deg270;
                QVector2D circPt = QVector2D(circRad * cosf(interp), circRad * sinf(interp) + circRad);

                m_samples.push_back(circPt);

            }           

        }

    }
    else { //a clothoid

        const float B = ComputeB();
        const float t1 = ComputeT1(B);
        const float t2 = ComputeT2(B);

        QVector2D basePt = Fresnel(t1) * M_PI * B;

        for (int i=0; i<nSamples; ++i) {

            float interp = float(i) / float(nSamples - 1);
            float eacht = t1 * (1.0f - interp) + t2 * interp;

            //get the point, translate
            QVector2D eachPt = Fresnel(eacht) * M_PI * B - basePt;

            //rotate
            QVector2D rotPt;

            float angle = -(t1 * t1);

            if (t2 < t1) {
                angle = -angle;
                eachPt.setX(-eachPt.x());
            }

            const float cosA = cosf(angle);
            const float sinA = sinf(angle);
            rotPt.setX(eachPt.x() * cosA - eachPt.y() * sinA);
            rotPt.setY(eachPt.x() * sinA + eachPt.y() * cosA);

            //qDebug() << "rotation of " << angle << " radians or degrees " << angle * 180.0f / PI << "based on t1 " << t1;

            m_samples.push_back(rotPt);

        }

    }

}

QVector2D Clothoid::Fresnel(const float s)
{

    const float t = fabs(s) / CURV_FACTOR;

    const float R = ((0.506f * t) + 1.0f) / ((1.79f * t * t) + (2.054f * t) + SQRT_2);
    const float A = 1.0f / ((0.803f * t * t * t) + (1.886f * t * t) + (2.524f * t) + 2.0f);

    const float cosVal = 0.5f - R * sinf(0.5f * M_PI * (A - t * t));
    const float sinVal = 0.5f - R * cosf(0.5f * M_PI * (A - t * t));

    if (s >= 0.0f)
        return QVector2D(cosVal, sinVal);
    else
        return QVector2D(-cosVal, -sinVal);

}

void Clothoid::SetArcLen(const float f)
{
    m_arcLen = qMax(f, 0.0f);
}

void Clothoid::SetStartCurv(const float f)
{
    m_startCurv = f;
}

void Clothoid::SetEndCurv(const float f)
{
    m_endCurv = f;
}

float Clothoid::ArcLen() const
{
    return m_arcLen;
}

float Clothoid::StartCurv() const
{
    return m_startCurv;
}

float Clothoid::EndCurv() const
{
    return m_endCurv;
}

float Clothoid::Curvature(float t) const
{
    const float interp = t / m_arcLen;
    return m_endCurv * interp + m_startCurv * (1.0f - interp);
}
