#include "clothoidspline.h"

ClothoidSpline::ClothoidSpline() :
    m_col1(0.0f, 0.0f, 1.0f),
    m_col2(0.0f, 1.0f, 0.0f),
    m_startPos(0.0f, 0.0f),
    m_startTan(0.0f),
    m_length(0.0f)
{
}

void ClothoidSpline::Add()
{

    if (m_clothoids.empty()) {
        Clothoid c;
        m_clothoids.push_back(c);

        m_length += c.ArcLen();
    }
    else {
        Clothoid c;
        c.SetStartCurv(m_clothoids.last().EndCurv());
        m_clothoids.push_back(c);

        m_length += c.ArcLen();
    }

    m_clothoids.last().DoSamples();

}

void ClothoidSpline::Add(ClothoidSpline & cs)
{

    m_clothoids += cs.m_clothoids;
    m_length += cs.m_length;

}

void ClothoidSpline::Add(Clothoid & c)
{

    m_clothoids.push_back(c);
    m_length += c.ArcLen();

}

void ClothoidSpline::Clear()
{
    m_clothoids.clear();
    m_length = 0.0f;
}

Clothoid & ClothoidSpline::operator[](const int ind)
{
    return m_clothoids[ind];
}

const Clothoid & ClothoidSpline::operator[](const int ind) const
{
    return m_clothoids[ind];
}

QVector2D ClothoidSpline::Rotate(const QVector2D & pt, const float angle)
{
    return QVector2D(pt.x() * cosf(angle) + pt.y() * (-sinf(angle)),
                     pt.x() * sinf(angle) + pt.y() * cosf(angle));
}

float ClothoidSpline::CanonEndTangent() const
{
    return CanonEndTangent(m_clothoids.size() - 1);
}

float ClothoidSpline::CanonEndTangent(const int ind) const
{

    float curAngle = 0.0f;

    for (int i=0; i<=ind; ++i)
        curAngle += m_clothoids[i].EndTanAngle();

    return curAngle;

}

QVector2D ClothoidSpline::CanonEndPosition() const
{
    return CanonEndPosition(m_clothoids.size() - 1);
}

QVector2D ClothoidSpline::CanonEndPosition(const int ind) const
{

    QVector2D pt(0.0f, 0.0f);
    float curAngle = 0.0f;

    for (int i=0; i<=ind; ++i) {
        pt += Rotate(m_clothoids[i].EndPoint(), curAngle);
        curAngle += m_clothoids[i].EndTanAngle();
    }

    return pt;

}

QVector2D ClothoidSpline::Position(const float t, float & r) const
{

    QVector2D pt(0.0f, 0.0f);
    r = 0.0f;
    float curarclen = 0.0f;

    int i;
    for (i=0; i<m_clothoids.size(); ++i) {

        if (curarclen + m_clothoids[i].ArcLen() > t)
            break;

        pt += Rotate(m_clothoids[i].EndPoint(), r);
        r += m_clothoids[i].EndTanAngle();
        curarclen += m_clothoids[i].ArcLen();

    }

    pt += Rotate(m_clothoids[i].Sample(t - curarclen), r);
    return pt;

}

QVector2D ClothoidSpline::EndPosition(const int ind) const
{
    return Rotate(CanonEndPosition(ind), m_startTan) + m_startPos;
}

QVector2D ClothoidSpline::EndPosition() const
{
    return Rotate(CanonEndPosition(), m_startTan) + m_startPos;
}

float ClothoidSpline::EndTangent(const int ind) const
{
    return CanonEndTangent(ind) + m_startTan;
}

float ClothoidSpline::EndTangent() const
{
    return CanonEndTangent() + m_startTan;
}

void ClothoidSpline::StartTrail(const int numc, ClothoidSpline & newcs) const
{

    newcs.Clear();

    const Clothoid & c = m_clothoids.first();

    //first assemble clothoid segments
    for (int i=-numc; i<0; ++i) {

        Clothoid newc;

        newc.SetArcLen(1.0f);
        newc.SetStartCurv(c.Curvature(float(i)));
        newc.SetEndCurv(c.Curvature(float(i+1)));

        newc.DoSamples();

        newcs.Add(newc);
    }

    //we now need to align this trailing curve such that where
    //it ends is where this curve begins
    float newtan = m_startTan - newcs.CanonEndTangent();
    QVector2D newpos = m_startPos - Rotate(newcs.CanonEndPosition(), newtan);

    newcs.SetStartPosition(newpos);
    newcs.SetStartTangent(newtan);

}

void ClothoidSpline::EndTrail(const int numc, ClothoidSpline & newcs) const
{

    newcs.Clear();

    const Clothoid & c = m_clothoids.last();

    //first assemble clothoid segments
    for (int i=0; i<numc; ++i) {

        Clothoid newc;

        newc.SetArcLen(1.0f);
        newc.SetStartCurv(c.Curvature(float(i)));
        newc.SetEndCurv(c.Curvature(float(i+1)));

        newc.DoSamples();

        newcs.Add(newc);
    }

    //we now need to align this trailing curve such that where
    //it ends is where this curve begins
    float newtan = EndTangent();
    QVector2D newpos = EndPosition();

    newcs.SetStartPosition(newpos);
    newcs.SetStartTangent(newtan);

}

float ClothoidSpline::AlignToEndpoints(const QVector2D & p1, const QVector2D & p2)
{

    const QVector2D p2m1 = p2 - p1;
    const QVector2D p2m1n = p2m1.normalized();
    const float ptan = atan2(p2m1.y(), p2m1.x());

    const QVector2D cp = CanonEndPosition();
    const float cptan = atan2(cp.y(), cp.x());

    const float error = p2m1.length() - cp.length();
    SetStartPosition(p1 + p2m1n * error / 2.0f);
    SetStartTangent(ptan - cptan);

    return fabs(error);

}

void ClothoidSpline::FromPolyline(const PolyLine & p, const bool forceFit)
{

    Clear();

    p.ArcLengths(m_lens);    
    p.Curvatures(m_curv);

    m_length = m_lens.last();

    //compute a polyline transformed to canonical position
    QVector2D startpt = p[0];
    QVector2D starttan = p.StartTangent();
    const float tanAngle = atan2f(starttan.y(), starttan.x());

    PolyLine p2;
    for (int i=0; i<p.Size(); ++i)
        p2.Add(Rotate(p[i] - startpt, -tanAngle));

    for (int i=0; i<p.Size()-1; ++i) {

        Clothoid c;

        c.SetArcLen(m_lens[i+1] - m_lens[i]);
        c.SetStartCurv(m_curv[i]);
        c.SetEndCurv(m_curv[i+1]);

        c.DoSamples();

        m_clothoids.push_back(c);

        /*
         * We perform a greedy-like local optimization to make the clothoid curve follow the
         * input polyline closely in space.  We do this by changing the curvature of this
         * clothoid's start/last clothoid's end curvature
         *
         * Need clothoid api to be good and efficient if this will work fast
         */

        if (forceFit && i > 0) {

            //determine if endpoint is within 2px of endpoint of polyline
            int numIter = 10;
            const float curvincr = 0.001f;

            QVector2D cpt = CanonEndPosition(i);
            float curlen = (p2[i+1] - cpt).length();

            Clothoid & c1 = m_clothoids[i-1];
            Clothoid & c2 = m_clothoids[i];

            while (numIter > 0 && curlen > 1.0f) {

                //by scaling curvature increment with iteration, we perform annealing
                float eachincr = curvincr * float(numIter);
                //float eachincr = curvincr;

                //try lowering
                c1.SetEndCurv(c1.EndCurv() - eachincr);
                c2.SetStartCurv(c2.StartCurv() - eachincr);
                c1.DoSamples();
                c2.DoSamples();
                QVector2D mpt = CanonEndPosition(i);
                float mlen = (p2[i+1] - mpt).length();

                //try increasing
                c1.SetEndCurv(c1.EndCurv() + eachincr * 2.0f);
                c2.SetStartCurv(c2.StartCurv() + eachincr * 2.0f);
                c1.DoSamples();
                c2.DoSamples();
                QVector2D ppt = CanonEndPosition(i);
                float plen = (p2[i+1] - ppt).length();

                //return to normal
                c1.SetEndCurv(c1.EndCurv() - eachincr);
                c2.SetStartCurv(c2.StartCurv() - eachincr);
                c1.DoSamples();
                c2.DoSamples();

                //qDebug() << "p2[i+1] " << p2[i+1] << "cpt " << cpt << "mpt " << mpt << "ppt " << ppt;

                if (mlen < curlen && mlen < plen) { //minus best
                    c1.SetEndCurv(c1.EndCurv() - eachincr);
                    c2.SetStartCurv(c2.StartCurv() - eachincr);
                    c1.DoSamples();
                    c2.DoSamples();

                    curlen = mlen;
                    cpt = mpt;
                }
                else if (plen < curlen && plen < mlen) { //plus best
                    c1.SetEndCurv(c1.EndCurv() + eachincr);
                    c2.SetStartCurv(c2.StartCurv() + eachincr);
                    c1.DoSamples();
                    c2.DoSamples();

                    curlen = plen;
                    cpt = ppt;
                } //else no change...

                --numIter;

            }            
        }

    }   

}

void ClothoidSpline::CurvatureSamples(const float incr, QVector <float> & curv) const
{

    curv.clear();

    int curi = 0;
    int curf = 0.0f;

    //sample curvature of this spline at every "incr" units
    for (float f=0.0f; f<m_length; f += incr, curf += incr) {

        //move to next clothoid if needed
        while (curf > m_clothoids[curi].ArcLen() && curi+1 < m_clothoids.size()) {           

            curf -= m_clothoids[curi].ArcLen();
            ++curi;

        }

        //get curvature at *local* arc length position of clothoid
        curv.push_back(m_clothoids[curi].Curvature(curf));        

    }

}

void ClothoidSpline::Remove()
{

    if (!m_clothoids.empty()) {
        m_length -= m_clothoids.last().ArcLen();
        m_clothoids.pop_back();
    }

}

int ClothoidSpline::Size() const
{
    return m_clothoids.size();
}

float ClothoidSpline::Length() const
{

    return m_length;

}

void ClothoidSpline::DrawColours(const QVector3D col1, const QVector3D col2)
{
    m_col1 = col1;
    m_col2 = col2;
}

void ClothoidSpline::DrawArcLengthGL(const float s1, const float s2) const
{

    if (m_clothoids.empty())
        return;

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();

    glTranslatef(m_startPos.x(), m_startPos.y(), 0);
    glRotatef(m_startTan * 180.0f / M_PI, 0, 0, 1);

    float scur = 0.0f;

    for (int i=0; i<m_clothoids.size(); ++i) {

        if (scur + m_clothoids[i].ArcLen() >= s1 && scur <= s2) {

            float start;
            float end;

            if (s1 > scur)
                start = s1 - scur;
            else
                start = 0.0f;

            if (s2 < scur + m_clothoids[i].ArcLen())
                end = s2 - scur;
            else
                end = m_clothoids[i].ArcLen();

            m_clothoids[i].DrawArcLengthGL(start, end);

        }

        QVector2D transPt = m_clothoids[i].EndPoint();
        glTranslatef(transPt.x(), transPt.y(), 0.0f);

        float rotPt = m_clothoids[i].EndTanAngle();
        glRotatef(rotPt * 180.0f / M_PI, 0, 0, 1);

        scur += m_clothoids[i].ArcLen();

    }

    glPopMatrix();

}

void ClothoidSpline::DrawGL(const bool closed) const
{
    DrawGL(0, m_clothoids.size(), closed);
}

void ClothoidSpline::DrawGL(const int start, const int end, const bool closed) const
{

    if (m_clothoids.empty())
        return;

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();

    glTranslatef(m_startPos.x(), m_startPos.y(), 0);
    glRotatef(m_startTan * 180.0f / M_PI, 0, 0, 1);

    for (int i=0; i<start; ++i) {
        QVector2D transPt = m_clothoids[i].EndPoint();
        glTranslatef(transPt.x(), transPt.y(), 0.0f);

        float rotPt = m_clothoids[i].EndTanAngle();
        glRotatef(rotPt * 180.0f / M_PI, 0, 0, 1);
    }

    for (int i=start; i<end; ++i) {

        if (i%2 == 0)
            glColor3f(m_col1.x(), m_col1.y(), m_col1.z());
        else
            glColor3f(m_col2.x(), m_col2.y(), m_col2.z());       

        m_clothoids[i].DrawGL();

        QVector2D transPt = m_clothoids[i].EndPoint();
        glTranslatef(transPt.x(), transPt.y(), 0.0f);

        float rotPt = m_clothoids[i].EndTanAngle();
        glRotatef(rotPt * 180.0f / M_PI, 0, 0, 1);

    }   

    glPopMatrix();

    if (closed) {

        const QVector2D v1 = EndPosition();
        const QVector2D v2 = StartPosition();

        glBegin(GL_LINES);
        glVertex2f(v1.x(), v1.y());
        glVertex2f(v2.x(), v2.y());
        glEnd();

    }

}

void ClothoidSpline::DrawTanSpaceGL(const int viewx, const int viewy, const int vieww, const int viewh, const float arclen) const
{

    //temporarily modify opengl viewport/projection matrix
    const double range = M_PI * 2.0f;
    Util::PushDrawOrthoGL(viewx, viewy, vieww, viewh, 0.0, qMax(arclen, 1.0f), -range, range);

    //axis lines
    glColor3f(0, 0, 0);
    glLineWidth(2.0f);

    glBegin(GL_LINES);

    glVertex2f(0.1f, 0);
    glVertex2f(0.1f, range);

    glVertex2f(0, 0);
    glVertex2f(m_length, 0);

    glEnd();

    float tanangle = 0.0f;
    glBegin(GL_LINE_STRIP);
    for (int i=0; i<m_clothoids.size(); ++i) {
        glVertex2f(m_lens[i], tanangle);

        tanangle += m_clothoids[i].EndTanAngle();

    }
    glEnd();

    glLineWidth(1.0f);

    //restore viewport, etc.
    Util::PopDrawOrthoGL();
}

void ClothoidSpline::DrawCurvSpaceGL(const int viewx, const int viewy, const int vieww, const int viewh, const float arclen) const
{

    if (m_clothoids.empty())
        return;

    //temporarily modify opengl viewport/projection matrix
    const double curvRange = 0.05f;
    Util::PushDrawOrthoGL(viewx, viewy, vieww, viewh, 0.0, qMax(arclen, 1.0f), -curvRange, curvRange * 2.0);

    //axis lines
    glColor3f(0, 0, 0);
    glLineWidth(2.0f);

    glBegin(GL_LINES);

    glVertex2f(0.1f, -curvRange);
    glVertex2f(0.1f, curvRange);

    glVertex2f(0, 0);
    glVertex2f(m_length, 0);

    float curArc = 0.0f;
    for (int i=0; i<m_clothoids.size(); ++i) {

        if (i%2 == 0)
            glColor3f(m_col1.x(), m_col1.y(), m_col1.z());
        else
            glColor3f(m_col2.x(), m_col2.y(), m_col2.z());

        glVertex2f(curArc, m_clothoids[i].StartCurv());
        curArc += m_clothoids[i].ArcLen();
        glVertex2f(curArc, m_clothoids[i].EndCurv());

    }   

    glEnd();

    glLineWidth(1.0f);

    Util::PopDrawOrthoGL();

}

QVector2D ClothoidSpline::StartPosition() const
{
    return m_startPos;
}

void ClothoidSpline::SetStartPosition(const QVector2D & p)
{
    m_startPos = p;
}

void ClothoidSpline::SetStartTangent(const QVector2D & t)
{
    m_startTan = atan2f(t.y(), t.x());
}

void ClothoidSpline::SetStartTangent(const float t)
{
    m_startTan = t;
}

float ClothoidSpline::StartTangent() const
{
    return m_startTan;
}


float ClothoidSpline::InterpolateG2(float f)
{
    return f * f * f * ( 6 * f * f - 15 * f + 10);
}

void ClothoidSpline::Blend(const ClothoidSpline &cs, const int blendwin, QVector <QVector2D> & blendpts) const
{

    blendpts.clear();

    const int range = qMin(qMin(blendwin, Size() / 2), cs.Size() / 2);

    ClothoidSpline blend1;
    EndTrail(range, blend1);

    ClothoidSpline blend2;
    cs.StartTrail(range, blend2);

    for (int j=0; j<range; ++j) {

        const float f = float(j) / float(range-1) * 0.5f;
        const float interp = InterpolateG2(f);

        QVector2D pt1 = EndPosition(Size()-1 - range + j);
        QVector2D pt2 = blend2.EndPosition(j);

        QVector2D pt = pt1 * (1.0f - interp) + pt2 * interp;

        blendpts.push_back(pt);

    }

    //the second "half" (between 0.5 - 1, we use endtrail)
    for (int j=0; j<range; ++j) {

        const float f = float(j) / float(range-1) * 0.5f + 0.5f;
        const float interp = InterpolateG2(f);

        QVector2D pt1 = blend1.EndPosition(j);
        QVector2D pt2 = cs.EndPosition(j);

        QVector2D pt = pt1 * (1.0f - interp) + pt2 * interp;

        blendpts.push_back(pt);

    }

}
